/*************************************************************************
 *
 *    Used with ICCARM and AARM.
 *
 *    (c) Copyright IAR Systems 2006
 *
 *    File name   : audio_class.c
 *    Description : AUDIO CLASS module
 *
 *    History :
 *    1. Data        : February 28, 2006
 *       Author      : Stanimir Bonev
 *       Description : Create
 *
 *    $Revision: 1.0 $
**************************************************************************/
#include "audio_class.h"

volatile Boolean SempEna,MicEna;
Int32U  SempCount,SempPeriod,Delta,MicCurrBuffer;
pInt16S pSpkData,pMicData;
Int32U  SempPerCurrHold,DeltaPer,MicSempCount;

union _Val
{
  Int32S Data;
  struct
  {
    Int16U DataLo;
    Int16S DataHi;
  };
} Val = {0x02000000};


const Int32U VolumeMul [] =
{
  // 256 - 0
  //-48dB- 0x0
  261,
  //-47dB- 0x1
  293,
  //-46dB- 0x2
  328,
  //-45dB- 0x3
  369,
  //-44dB- 0x4
  414,
  //-43dB- 0x5
  464,
  //-42dB- 0x6
  521,
  //-41dB- 0x7
  584,
  //-40dB- 0x8
  655,
  //-39dB- 0x9
  735,
  //-38dB- 0xA
  825,
  //-37dB- 0xB
  926,
  //-36dB- 0xC
  1039,
  //-35dB- 0xD
  1165,
  //-34dB- 0xE
  1308,
  //-33dB- 0xF
  1467,
  //-32dB- 0x10
  1646,
  //-31dB- 0x11
  1847,
  //-30dB- 0x12
  2072,
  //-29dB- 0x13
  2325,
  //-28dB- 0x14
  2609,
  //-27dB- 0x15
  2927,
  //-26dB- 0x16
  3285,
  //-25dB- 0x17
  3685,
  //-24dB- 0x18
  4135,
  //-23dB- 0x19
  4640,
  //-22dB- 0x1A
  5206,
  //-21dB- 0x1B
  5841,
  //-20dB- 0x1C
  6554,
  //-19dB- 0x1D
  7353,
  //-18dB- 0x1E
  8250,
  //-17dB- 0x1F
  9257,
  //-16dB- 0x20
  10387,
  //-15dB- 0x21
  11654,
  //-14dB- 0x22
  13076,
  //-13dB- 0x23
  14672,
  //-12dB- 0x24
  16462,
  //-11dB- 0x25
  18471,
  //-10dB- 0x26
  20724,
  //-9dB - 0x27
  23253,
  //-8dB - 0x28
  26090,
  //-7dB - 0x29
  29274,
  //-6dB - 0x2A
  32846,
  //-5dB - 0x2B
  36854,
  //-4dB - 0x2C
  41350,
  //-3dB - 0x2D
  46396,
  //-2dB - 0x2E
  52057,
  //-1dB - 0x2F
  58409,
  // 0dB - 0x30
  65536,
  // 1dB - 0x31
  73533,
  // 2dB - 0x32
  82505,
  // 3dB - 0x33
  92572,
  // 4dB - 0x34
  103868,
  // 5dB - 0x35
  116541,
  // 6dB - 0x36
  130762,
};

#pragma data_alignment=4
__no_init Int16S AudioSpkData[SampRerFrame * 3];

#pragma data_alignment=4
__no_init Int16S AudioMicData1[SampRerFrame * 2];
#pragma data_alignment=4
__no_init Int16S AudioMicData2[SampRerFrame * 2];

#pragma data_alignment=4
Int8U AudioBuf[2];

Int8U   AudioRequest,AudioCS,AudioCN,AudioId;
Int16U  AudioDataSize;

Int16S  AudioFeat1Vol;
Int32U  AudioSpkVolMul;
Boolean AudioFeat1Mute;

Int16S  AudioFeat2Vol;
Int32U  AudioMicVolMul;
Boolean AudioFeat2Mute;


/*************************************************************************
 * Function Name: AudioClassInit
 * Parameters: VicIrqSlots_t IntrSlot
 *
 * Return: none
 *
 * Description: USB Class Audio Init
 *
 *************************************************************************/
void AudioClassInit (VicIrqSlots_t IntrSlot)
{
  // Init Audio Class variables
  SempEna         =\
  AudioFeat1Mute  =\
  AudioFeat2Mute  = FALSE;
  SempCount       =\
  AudioFeat1Vol   =\
  AudioFeat2Vol   = 0;
  pSpkData        = AudioSpkData;
  SempPerCurrHold = SempPeriod;

  // Registered The Class Request
  UsbCoreUserFuncRegistered(UsbClassAudioRequest,UsbUserClass);
  // Registered User Ep0 Data receive
  UsbCoreUserFuncRegistered(UsbClassAudioData,UsbClassEp0OutPacket);

  // PWM DAC (TIM3) 10 bit
//  SCU_PRR1_bit.RST_GPIO4 = 0;       // Reset GPIO4
//  SCU_PCGR1_bit.GPIO4    = 1;       // GPIO4 Port clock runing
  SCU_PRR1_bit.RST_GPIO4 = 1;       // GPIO4 out from reset
  SCU_GPIOOUT4_bit.P4OUT6 = 2;      // GPIO4.6 Alternate Output 2 (Out comp/PWM)
  SCU_GPIOANA_bit.P46A = 0;      // GPIO4.6 Alternate Output 2 (Out comp/PWM)

  SCU_PRR1_bit.RST_TIM23 = 0;       // Reset of TIM23
  SCU_PCGR1_bit.TIM23 = 1;          // Enable clock of TIM23
  SCU_PRR1_bit.RST_TIM23 = 1;       // Release reset of TIM23
  TIM3_CR2 = 0;                     // Set TIM3 perscaler (max resolution)
  TIM3_CR1_bit.ECKEN = 0;           // TIM clock from Internal clock, divided by CC[7:0] prescaler
  TIM3_CR1_bit.OLVL1 = 0;           // Level to be applied to the OCMP1 by OC1R comparison
  TIM3_CR1_bit.OLVL2 = 1;           // Level to be applied to the OCMP1 by OC2R comparison
  TIM3_CR1_bit.PWM = 1;             // PWM mode enable
  TIM3_CR1_bit.OC1E = 1;            // Enable mach pin action
  TIM3_OCR2 = 0x3FF;                // PWM period
  TIM3_OCR1 = 0x200;                // PWM pulse
  TIM3_CR1_bit.EN = 1;              // enable counting

  VIC_Config(TIM3_ITLine,VIC_FIQ,15);
  VIC_ITCmd(TIM3_ITLine,ENABLE);

  SCU_APBPeriphClockConfig(__ADC, ENABLE);    /* Enable the clock for the ADC */
  ADC_DeInit();                               /* ADC Deinitialization */

  // ADC init
  // Assign P4.7 to ADC ch 7
  SCU_PRR1_bit.RST_GPIO4= 0;        // Reset GPIO9
//  SCU_PCGR1_bit.GPIO4   = 1;        // GPIO9 Port clock runing
  SCU_PRR1_bit.RST_GPIO4= 1;        // GPIO9 out from reset
  SCU_GPIOANA_bit.P47A = 1;         // Analog mode on
  SCU_GPIOIN4_bit.P4IN7 = 0;        // On-chip peripheral disconnected
  SCU_GPIOOUT4_bit.P4OUT7 = 0;      // On-chip peripheral disconnected
  SCU_PRR1_bit.RST_ADC = 0;         // Reset of ADC
//  SCU_PCGR1_bit.ADC = 1;            // Enable clock of ADC
  SCU_PRR1_bit.RST_ADC = 1;         // Release reset of ADC
  ADC_CR_bit.POR  = 1;              // Power on mode
  ADC_CR_bit.STB  = 0;              // The analog block is kept powered on
  ADC_CR_bit.SCE  = 0;              // Single channel mode
  ADC_CR_bit.SC   = 7;              // Channel 7
  ADC_CR_bit.AWDI = 0;              // AWD interrupt disabled
  ADC_CR_bit.ECVI = 0;              // End of conversion interrupt disable
  ADC_CCR_bit.CC7 = 3;              // A/D Conversion without watchdog feature on channel
  ADC_PRS = 16;                     // PCLK/16 = 3MHz

  // Init Tick Timer - TIM0
  TIM1_CR2 = 0;                     // Set TIM01 perscaler (max resolution)
  TIM1_CR1_bit.ECKEN = 0;           // TIM clock from Internal clock, divided by CC[7:0] prescaler
  TIM1_CR2_bit.OC1IE = 1;           // TIM0 output compare interrupt enable
  // Set timer period
  SempPeriod = (SCU_GetPCLKFreqValue()*1000)/SampFreq;
  DeltaPer = SempPeriod/(SampRerFrame*2);
  TIM1_OCR1 = TIM1_CNTR + SempPeriod;
  TIM1_CR1_bit.EN = 1;              // enable counting
  // Set interrupt
  VIC_Config(TIM1_ITLine,VIC_IRQ , IntrSlot);
  VIC_ITCmd(TIM1_ITLine, ENABLE);

}

/*************************************************************************
 * Function Name: UsbClassAudioConfigure
 * Parameters:  void * pArg
 *
 * Return: void *
 *
 * Description: USB Class Audio configure
 *
 *************************************************************************/
void * UsbClassAudioConfigure (void * pArg)
{
UsbDevCtrl_t * pUsbDevCtrl = (UsbDevCtrl_t *)pArg;
  if(pUsbDevCtrl == NULL)
  {
    if(UsbCoreReq(UsbCoreReqConfiquration) != 0)
    {
      MicEna = FALSE;
      // disable all class EPs
      USB_UserFuncRegistered(NULL,SpkEp);
      USB_RealizeEndPoint((USB_Endpoint_t)SpkEp,
                           1,
                           0,
                           FALSE,
                           EP_ISO,
                           SpkEpSlot);
      USB_UserFuncRegistered(NULL,MicEp);
      USB_RealizeEndPoint((USB_Endpoint_t)MicEp,
                           1,
                           0,
                           FALSE,
                           EP_ISO,
                           MicEpSlot);
    }
  }
  else
  {
    if((pUsbDevCtrl->Configuration == 1))
    {
      if(pUsbDevCtrl->Interface == 1)
      {
        if(pUsbDevCtrl->AlternateSetting == 0)
        {
          USB_UserFuncRegistered(NULL,SpkEp);
          USB_RealizeEndPoint((USB_Endpoint_t)SpkEp,
                               1,
                               0,
                               FALSE,
                               EP_ISO,
                               SpkEpSlot);
        }
        else if (pUsbDevCtrl->AlternateSetting == 1)
        {
          USB_UserFuncRegistered(AudioInHadler,SpkEp);
          USB_RealizeEndPoint((USB_Endpoint_t)SpkEp,
                               1,
                               SpkEpMaxSize,
                               TRUE,
                               EP_ISO,
                               SpkEpSlot);
        }
      }
      else if (pUsbDevCtrl->Interface == 2)
      {
        if(pUsbDevCtrl->AlternateSetting == 0)
        {
          MicEna = FALSE;
          USB_UserFuncRegistered(NULL,UsbSofEvent);
          USB_RealizeEndPoint((USB_Endpoint_t)MicEp,
                               1,
                               0,
                               FALSE,
                               EP_ISO,
                               MicEpSlot);
        }
        else if (pUsbDevCtrl->AlternateSetting == 1)
        {
          USB_UserFuncRegistered(AudioOutHadler,MicEp);
          USB_RealizeEndPoint((USB_Endpoint_t)MicEp,
                               1,
                               MicEpMaxSize,
                               TRUE,
                               EP_ISO,
                               MicEpSlot);
          pMicData = AudioMicData1;
          MicCurrBuffer = 0;
        }
      }
    }
  }
  return(NULL);
}

/*************************************************************************
 * Function Name: AudioInHadler
 * Parameters:  void * pArg
 *
 * Return: none
 *
 * Description: USB Class Audio Out EP handler
 *
 *************************************************************************/
void AudioInHadler (void *Arg)
{
Int32U ReceievedBytes = SpkEpMaxSize;
volatile Int32U Status;
  Status = USB_EpRead((USB_Endpoint_t)SpkEp,
            (Int32U *)pSpkData,
            &ReceievedBytes);
  if(Status != USB_OK)
  {
    Status = Status;
  }
  if(ReceievedBytes)
  {
    pSpkData += ReceievedBytes/sizeof(Int16U);
    if(pSpkData >= AudioSpkData + sizeof(AudioSpkData)/sizeof(Int16U))
    {
      pSpkData = AudioSpkData;
    }

    if(SempEna)
    {
      // Adapt sample rate
      if (Delta > (SampRerFrame * SubFrameSize))
      {
        SempPerCurrHold = SempPeriod - DeltaPer;
      }
      else if (Delta < (SampRerFrame * SubFrameSize))
      {
        SempPerCurrHold = SempPeriod + DeltaPer;
      }
      else
      {
        SempPerCurrHold = SempPeriod;
      }
    }
    else
    {
      if(pSpkData >= (AudioSpkData + (2*sizeof(AudioSpkData))/(3*sizeof(Int16U))))
      {
        // Enable output
        SempCount = 0;
        SempEna   = TRUE;
      }
    }
    Delta  += ReceievedBytes;
  }
}

/*************************************************************************
 * Function Name: AudioOutHadler
 * Parameters:  void * pArg
 *
 * Return: none
 *
 * Description: USB Class Audio In EP handler
 *
 *************************************************************************/
void AudioOutHadler (void *Arg)
{
Int32U PacketLength;
  MicEna = TRUE;
  // Set pointer to free buffer
  if(++MicCurrBuffer & 1)
  {
    PacketLength = (Int32U)pMicData - (Int32U)AudioMicData1;
    pMicData = AudioMicData2;
  }
  else
  {
    PacketLength = (Int32U)pMicData - (Int32U)AudioMicData2;
    pMicData = AudioMicData1;
  }

  // Send microphone output buffer
  USB_EpWrite((USB_Endpoint_t)MicEp,
              (pInt32U)((MicCurrBuffer & 1)?AudioMicData1:AudioMicData2),
              &PacketLength);
}

/*************************************************************************
 * Function Name: TIM0_ISR
 * Parameters: none
 *
 * Return: none
 *
 * Description:
 *		
 *************************************************************************/
void TIM1_ISR (void)
{
static union _Val MicTemp;

  Val.DataHi = ADC_DR7_bit.CDATA;

 // Val.DataHi += 20;
//  if(Val.DataHi > 1023)
//  {
//    Val.DataHi = 0;
//  }

  TIM3_SR  &= ~(1UL << 11);
  TIM3_CR2 |=  (1UL << 11);

  // Start conversion
  ADC_CR_bit.STR = 1;

/*
  if (SempEna)
  {
    if(Delta > 1)
    {
      // Get next input sample
      Val.Data = AudioSpkData[SempCount++];
      Delta   -= 2;
      // Correct pointer to sample buffer
      if(SempCount >= sizeof(AudioSpkData)/sizeof(Int16U))
      {
        SempCount = 0;
      }
      if(!AudioFeat1Mute && AudioSpkVolMul)
      {
        // Apply volume
        Val.Data   *= AudioSpkVolMul;
        // Add offset
        Val.DataHi += 0x200;
        // Check for overflow and correct value
        if(Val.DataHi < 0)
        {
          Val.DataHi = 0;
        }
        else if(Val.DataHi > 0x3FF)
        {
          Val.DataHi = 0x3FF;
        }
      }
      else
      {
        // set middle of range
        Val.DataHi = 0x200;
      }
    }
    else
    {
      // Disabe output stream after emptying the buffer
      SempEna = FALSE;
      SempPerCurrHold = SempPeriod;
    }
  }
  else
  {
    SempPerCurrHold = SempPeriod;
  }

  TIM3_SR  &= ~(1UL << 11);
  TIM3_CR2 |=  (1UL << 11);

  if(MicEna)
  {
    if(!AudioFeat2Mute && AudioMicVolMul)
    {
      // Get ADC sample and remove offset
      MicTemp.Data  = ADC_DR7_bit.CDATA - 0x200;
      // Start conversion
      ADC_CR_bit.STR = 1;
      // Add volume
      MicTemp.Data *= AudioMicVolMul;
      // check for overflow and correct the result
      if(MicTemp.DataHi < -511)
      {
        MicTemp.Data = -511 * 65536;
      }
      else if(MicTemp.DataHi > 511)
      {
        MicTemp.Data = 511 * 65536;
      }
      // Scale to 16bit
      MicTemp.Data <<= 6;
    }
    else
    {
      // set middle of range
      MicTemp.DataHi = 0;
    }
    // write to output buffer
    *pMicData++ = MicTemp.DataHi;
  }
  */
  TIM1_OCR1 += SempPeriod; // reload output compare
  TIM1_SR_bit.OCF1 = 0;         // clear interrupt flag
}

/*************************************************************************
 * Function Name: fiq_handler
 * Parameters: none
 *
 * Return: none
 *
 * Description:
 *		
 *************************************************************************/
void fiq_handler (void)
{
  // Set DAC
  TIM3_OCR1 = Val.DataHi;
  TIM3_CR2 &= ~(1UL << 11);
}

/*************************************************************************
 * Function Name: AudioFeatureGetReg
 * Parameters:  Int32U CS, Int32U Id
 *
 * Return: Boolean
 *
 * Description:
 *
 *************************************************************************/
static Boolean AudioFeatureGetReg (Int32U CS, Int32U Id)
{
  switch (CS)
  {
  case REQUEST_GET_CUR:
    // Load current value of Volume in the transmit buffer
    if (Id == FeatUnit1Id)
    {
      AudioBuf[1] = (AudioFeat1Vol>>8)& 0xFF;
      AudioBuf[0] =  AudioFeat1Vol    & 0xFF;
    }
    else if (Id == FeatUnit2Id)
    {
      AudioBuf[1] = (AudioFeat2Vol>>8)& 0xFF;
      AudioBuf[0] =  AudioFeat2Vol    & 0xFF;
    }
    else
    {
      return(FALSE);
    }
    break;
  case REQUEST_GET_MIN:
    // Load minimum value of Volume in the transmit buffer
    if (Id == FeatUnit1Id)
    {
      AudioBuf[1] = (Feat1MinVol>>8)  & 0xFF;
      AudioBuf[0] =  Feat1MinVol      & 0xFF;
    }
    else if (Id == FeatUnit2Id)
    {
      AudioBuf[1] = (Feat2MinVol>>8)  & 0xFF;
      AudioBuf[0] =  Feat2MinVol      & 0xFF;
    }
    else
    {
      return(FALSE);
    }
    break;
  case REQUEST_GET_MAX:
    // Load maximum value of Volume in the transmit buffer
    if (Id == FeatUnit1Id)
    {
      AudioBuf[1] = (Feat1MaxVol>>8)  & 0xFF;
      AudioBuf[0] =  Feat1MaxVol      & 0xFF;
    }
    else if (Id == FeatUnit2Id)
    {
      AudioBuf[1] = (Feat2MaxVol>>8)  & 0xFF;
      AudioBuf[0] =  Feat2MaxVol      & 0xFF;
    }
    else
    {
      return(FALSE);
    }
    break;
  case REQUEST_GET_RES:
    // Load resolution value of Volume in the transmit buffer
    if (Id == FeatUnit1Id)
    {
      AudioBuf[1] = (Feat1ResVol>>8)  & 0xFF;
      AudioBuf[0] =  Feat1ResVol      & 0xFF;
    }
    else if (Id == FeatUnit2Id)
    {
      AudioBuf[1] = (Feat2ResVol>>8)  & 0xFF;
      AudioBuf[0] =  Feat2ResVol      & 0xFF;
    }
    else
    {
      return(FALSE);
    }
    break;
  default:
    return(FALSE);
  }
  return(TRUE);
}

/*************************************************************************
 * Function Name: UsbClassAudioRequest
 * Parameters:  void * pArg
 *
 * Return: void *
 *
 * Description: USB Class Audio Requests
 *
 *************************************************************************/
void * UsbClassAudioRequest (void * pArg)
{
UsbEpCtrl_t * pAdioReqCtrl = (UsbEpCtrl_t *) pArg;
UsbSetupPacket_t * pAudioReqPacket = (UsbSetupPacket_t *)pAdioReqCtrl->pData;
  // Validate Request
  switch (pAudioReqPacket->mRequestType.Recipient)
  {
  case UsbRecipientInterface:
    // Feature Unit requests only imlement for interface 0
    if ((pAudioReqPacket->wIndex.Word == (FeatUnit1Id << 8)) ||
        (pAudioReqPacket->wIndex.Word == (FeatUnit2Id << 8)))
    {
      // Requset type
      switch (pAudioReqPacket->wValue.Hi)
      {
      case FU_MUTE_CONTROL:
        if ((pAudioReqPacket->bRequest == REQUEST_SET_CUR) &&
            (pAudioReqPacket->wLength.Word == 1))
        {
          // Set mute flag
          AudioRequest  = pAudioReqPacket->bRequest;
          AudioId       = pAudioReqPacket->wIndex.Hi;
          AudioCS       = pAudioReqPacket->wValue.Hi;
          AudioCN       = pAudioReqPacket->wValue.Lo;
          AudioDataSize = pAudioReqPacket->wLength.Word;
          pAdioReqCtrl->Counter= 1;
          pAdioReqCtrl->pData = AudioBuf;
          return((void *)UsbUserReceivePacket);
        }
        else if ((pAudioReqPacket->bRequest == REQUEST_GET_CUR) &&
                 (pAudioReqPacket->wLength.Word == 1))
        {
          // Read mute flag
          if(pAudioReqPacket->wIndex.Hi == FeatUnit1Id)
          {
            AudioBuf[0]  = AudioFeat1Mute;
          }
          else
          {
            AudioBuf[0]  = AudioFeat2Mute;
          }
          pAdioReqCtrl->Counter= 1;
          pAdioReqCtrl->pData = AudioBuf;
          return((void*)UsbUserSendPacket);
        }
        break;
      case FU_VOLUME_CONTROL:
        if(pAudioReqPacket->bRequest & 0x80)
        {
          // Read differnt volume valules
          if((pAudioReqPacket->wLength.Word == 2) &&
              AudioFeatureGetReg(pAudioReqPacket->bRequest,pAudioReqPacket->wIndex.Hi))
          {
            pAdioReqCtrl->Counter = 2;
            pAdioReqCtrl->pData = AudioBuf;
            return((void*)UsbUserSendPacket);
          }
        }
        else if((pAudioReqPacket->bRequest == REQUEST_SET_CUR) &&
                (pAudioReqPacket->wLength.Word  == 2))
        {
          // Set volume valule
          AudioRequest  = pAudioReqPacket->bRequest;
          AudioId       = pAudioReqPacket->wIndex.Hi;
          AudioCS       = pAudioReqPacket->wValue.Hi;
          AudioCN       = pAudioReqPacket->wValue.Lo;
          AudioDataSize = pAudioReqPacket->wLength.Word;
          pAdioReqCtrl->Counter = 2;
          pAdioReqCtrl->pData = AudioBuf;
          return((void *)UsbUserReceivePacket);
        }
        break;
      }
    }
    // Selector Unit requests only imlement for interface 0
    else if (pAudioReqPacket->wIndex.Word == (SelUnit1ID << 8))
    {
      if(pAudioReqPacket->wValue.Word == 0)
      {
        // Read different selector unit values
        if(pAudioReqPacket->bRequest & 0x80)
        {
          if(pAudioReqPacket->wLength.Word == 1)
          {
            AudioBuf[0] = 1;
            pAdioReqCtrl->Counter = 1;
            pAdioReqCtrl->pData = AudioBuf;
            return((void*)UsbUserSendPacket);
          }
        }
        else
        {
          // Set channel
          AudioRequest  = pAudioReqPacket->bRequest;
          AudioId       = pAudioReqPacket->wIndex.Hi;
          AudioDataSize = pAudioReqPacket->wLength.Word;
          pAdioReqCtrl->Counter = 1;
          pAdioReqCtrl->pData = AudioBuf;
          return((void *)UsbUserReceivePacket);
        }
      }
    }
    return((void *)UsbUserStallCtrlEp);
  case UsbRecipientEndpoint:
    return((void *)UsbUserStallCtrlEp);
  }
  return((void *)UsbUserStallCtrlEp);
}

/*************************************************************************
 * Function Name: UsbClassAudioData
 * Parameters:  void * pArg
 *
 * Return: void *
 *
 * Description: USB Class Audio Data receive
 *
 *************************************************************************/
void * UsbClassAudioData (void * pArg)
{
  if((AudioId == FeatUnit1Id) ||
     (AudioId == FeatUnit2Id))
  {
    switch (AudioCS)
    {
    case FU_MUTE_CONTROL:
      // Set mute flag
      if (AudioId == FeatUnit1Id)
      {
        AudioFeat1Mute = AudioBuf[0];
      }
      else if (AudioId == FeatUnit2Id)
      {
        AudioFeat2Mute = AudioBuf[0];
      }
      else
      {
        return((void*)UsbUserStallCtrlEp);
      }
      break;
    case FU_VOLUME_CONTROL:
      // Set volume value
      if (AudioId == FeatUnit1Id)
      {
        AudioFeat1Vol = AudioBuf[0] + ((Int16U)AudioBuf[1]<<8);
        if((Int16U)AudioFeat1Vol == MinVol)
        {
          AudioSpkVolMul = 0;
        }
        else
        {
          if (AudioFeat1Vol < (Int16S)Feat1MinVol)
          {
            AudioFeat1Vol = Feat1MinVol;
          }
          else if(AudioFeat1Vol > (Int16S)Feat1MaxVol)
          {
            AudioFeat1Vol = Feat1MaxVol;
          }
          // -48dB - +6dB multiply
          AudioSpkVolMul = (VolumeMul[((AudioFeat1Vol>>8)+0x30) & 0x3F]) >> 6;
        }
      }
      else if (AudioId == FeatUnit2Id)
      {
        AudioFeat2Vol = AudioBuf[0] + ((Int16U)AudioBuf[1]<<8);
        if((Int16U)AudioFeat2Vol == MinVol)
        {
          AudioMicVolMul = 0;
        }
        else
        {
          if (AudioFeat2Vol < (Int16S)Feat2MinVol)
          {
            AudioFeat2Vol = Feat2MinVol;
          }
          else if(AudioFeat2Vol > (Int16S)Feat2MaxVol)
          {
            AudioFeat2Vol = Feat2MaxVol;
          }
          // -48dB - +6dB multiply
          AudioMicVolMul = VolumeMul[((AudioFeat2Vol>>8)+0x30) & 0x3F];
        }
      }
      else
      {
        return((void*)UsbUserStallCtrlEp);
      }
      break;
    default:
      return((void*)UsbUserStallCtrlEp);
    }
  }
  else if (AudioId == SelUnit1ID)
  {
    // empty
  }
  else
  {
    return((void*)UsbUserStallCtrlEp);
  }
  return((void*)UsbUserSendAckn);
}
